/*
   IEEE_IoT_Sketch03_Standalone_V02

   Standalone Weather Station

   22 October 2016

   Hardware connections NodeMCU to Level Converter & LCD:
     Signal   NodeMCU       Lvl Conv      LCD
     ------   -------       --------      ---
     5V ----- VIN --------------- HV ---- VCC
     3.3V --- 3V3 --------- LV  |
     GND ---- GND --------- GND | GND --- GND
     SCL ---- D1 (GPIO5) -- LV1 | HV1 --- SCL
     SDA ---- D2 (GPIO4) -- LV2 | HV2 --- SDA

   Hardware connections from Level Converter to Sensors:
     Signal   Lvl Conv      BME280        BH1750
     -----    --------      ------        ------
     3.3V --- LV ---------- 3.3 --------- VCC
     GND ---- GND --------- GND --------- GND
     SCL ---- LV1 --------- SCL --------- SCL
     SDA ---- LV2 --------- SDA --------- SDA

   Units Switch connected between NodeMCU pin D7 (GPIO13) and GND

   Set serial monitor to 115,200 baud
*/

/* Libraries that can be installed from menu Sketch | Include Libraries | Manage Libraries...
      Liquidcrystal_I2C by Frank de Brabander
      BME280 by Tyler Glenn

   Libraries that must be downloaded are:
      BH1750
      RunningAverage
*/
#include <Wire.h>                // [part of IDE] I2C bus
#include <LiquidCrystal_I2C.h>   // https://github.com/fdebrabander/Arduino-LiquidCrystal-I2C-library
#include <BME280.h>              // [manager] https://github.com/finitespace/BME280
#include <BH1750.h>              // https://github.com/claws/BH1750
#include <RunningAverage.h>      // https://github.com/RobTillaart/Arduino/tree/master/libraries/RunningAverage

// **************** PROGRAM PARAMETERS *******************
const float STATION_ELEV = 125.0;               // Altitude of Montgomery College (m)
const int LCD_ADDR = 0x27;                      // backlight: yellow = 0x27, blue = 0x3F
const int UNITS_PIN = 13;                       // GPIO13 = NodeMCU pin D7
const long UPDATE_INTERVAL_SENSORS = 1000;      // 1 second update interval
const int SENSOR_SAMPLES = 20;                  // arbitrary number of samples to smooth
const long SPLASHSCREEN_DELAY = 2000;           // 2 second delay

// LEDs are connected but not used in this sketch
const int LED_D5 = 14;
const int LED_D6 = 12;

const float HPA_TO_INHG = 0.0295299830714;  // hPa to inHg pressure

// structure to hold sensor measurements
struct sensorData
{
  float stationPressure;
  float seaLevelPressure;
  float temperature;
  float humidity;
  unsigned int lightIntensity;  // other sensors may need float
  float voltage;
} rawData, smoothData; // declare struct variables

// ********* INSTANTIATE OBJECTS *************************
// usage: LiquidCrystal_I2C your_object_name(I2C_address, columns, rows);
// run IEEE_IoT_I2C_scanner if address is not known
LiquidCrystal_I2C lcd(LCD_ADDR, 16, 2);

// Create instances of the sensor objects
BME280 myBME280;        // barometric pressure / temperature / humidity sensor
BH1750 myBH1750;        // light intensity sensor

// instantiate objects to hold smoothed values
RunningAverage avgStationPressure(SENSOR_SAMPLES);
RunningAverage avgSeaLevelPressure(SENSOR_SAMPLES);
RunningAverage avgTemperature(SENSOR_SAMPLES);
RunningAverage avgHumidity(SENSOR_SAMPLES);

// *******************************************************
// ********************* SETUP ***************************
// *******************************************************
void setup()
{
  Serial.begin(115200);             // initialize the serial port
  lcd.begin();                      // initialize the lcd
  splashScreen();                   // show the splash screen
  myBME280.begin();                 // initialize BME280 pressure/temperature/humidity
  myBH1750.begin();                 // initialize BH1750 light sensor
  pinMode(UNITS_PIN, INPUT_PULLUP); // pull NodeMCU pin HIGH for units selection switch
// fix to keep LEDs dark - they are not used in this sketch
  pinMode(LED_D5, OUTPUT);
  pinMode(LED_D6, OUTPUT);
  digitalWrite(LED_D5, LOW);
  digitalWrite(LED_D6, LOW);
  } //setup()

// *******************************************************
// ******************** LOOP *****************************
// *******************************************************
void loop()
{
  rawData = readSensors();          // load all sensor data into rawData struct
  smoothData = averageSensorData(rawData);
  bool units = readUnits(UNITS_PIN);
  printToLCD(smoothData, units);
  printToSerialPort(rawData, smoothData, units);
  delay(UPDATE_INTERVAL_SENSORS);
} // loop()

// *******************************************************
// ******************* FUNCTIONS *************************
// *******************************************************

// read the GPIO pin used to set metric/Imperial
// leave pin open for metric, short to ground for Imperial
bool readUnits(int pin)
{
  if (digitalRead(pin) == HIGH)
  {
    return true;   // metric units
  }
  else
  {
    return false; // Imperial units
  }
} // readUnits()

// *******************************************************
// smooth the sensor data
sensorData averageSensorData(sensorData raw)
{
  sensorData smooth;  // temporary variable to hold smoothed data

  // average the sensor data
  avgStationPressure.addValue(raw.stationPressure);
  avgSeaLevelPressure.addValue(raw.seaLevelPressure);
  avgTemperature.addValue(raw.temperature);
  avgHumidity.addValue(raw.humidity);

  // fill the temporray struct with the averaged sensor data
  smooth.stationPressure = avgStationPressure.getAverage();
  smooth.seaLevelPressure = avgSeaLevelPressure.getAverage();
  smooth.temperature = avgTemperature.getAverage();
  smooth.humidity = avgHumidity.getAverage();
  smooth.lightIntensity = rawData.lightIntensity;  // don't average the light intensity
  return smooth;
} //averageSensorData

// *******************************************************
// show the splash screen for a set number of seconds
void splashScreen()
{
  lcd.backlight();                     // turn on backlight
  lcd.setCursor(0, 0);                 // upper left
  lcd.print("IEEE Northern VA");
  lcd.setCursor(1, 1);                 // indent next line to center text
  lcd.print("IoT Wx Station");
  delay(SPLASHSCREEN_DELAY);           // display for x seconds
  lcd.clear();                         // clear screen
} // splashScreen()

// *******************************************************
// read all sensors into a sensors struct
sensorData readSensors()
{
  sensorData temporaryData;  // temporary variable to hold readings

  // Pressure Parameters: (float& pressure, float& temp, float& humidity, bool metric, int pressureUnit)
  byte pressureUnit = 1; // unit: 0 = Pa, 1 = hPa, 2 = Hg, 3 = atm, 4 = bar, 5 = torr, 6 = N/m^2, 7 = psi

  // read each BME280 value individually for clarity of code
  // or read them all at once with the following line
  // myBME280.ReadData(temporaryData.stationPressure, temporaryData.temperature, temporaryData.humidity, true, pressureUnit);

  temporaryData.stationPressure = myBME280.ReadPressure(pressureUnit);
  temporaryData.temperature = myBME280.ReadTemperature();
  temporaryData.humidity = myBME280.ReadHumidity();

  // calculate the sealevel (relative) pressure
  temporaryData.seaLevelPressure = calculateSeaLevelPressure(temporaryData.temperature, temporaryData.stationPressure, STATION_ELEV);

  // read light level in lux
  temporaryData.lightIntensity = myBH1750.readLightLevel();

  // read analog voltage from the Analog to Digital Converter
  // on NodeMCU this is 0 - 1023 for voltages 0 to 3.3V
  temporaryData.voltage = 3.3 * analogRead(A0) / 1023.0;

  return temporaryData;
} // readSensors()

// *******************************************************
// print raw and smoothed data to the serial port
void printToSerialPort(sensorData dataRaw, sensorData dataSmooth, bool metricUnits)
{
  // convert display data to Imperial if requested by user
  if (metricUnits == false) {
    dataRaw.temperature = dataRaw.temperature * 1.8 + 32;
    dataRaw.stationPressure = dataRaw.stationPressure * HPA_TO_INHG;
    dataRaw.seaLevelPressure = dataRaw.seaLevelPressure * HPA_TO_INHG;
    dataSmooth.temperature = dataSmooth.temperature * 1.8 + 32;
    dataSmooth.stationPressure = dataSmooth.stationPressure * HPA_TO_INHG;
    dataSmooth.seaLevelPressure = dataSmooth.seaLevelPressure * HPA_TO_INHG;
  }

  // '\t' is the C++ escape sequence for tab
  // header line
  Serial.println("\tTemp:\tHumd:\tPsta:\tPsea:\tLux:\tVolt:");

  // first line - raw data
  Serial.print("RAW\t");
  Serial.print(dataRaw.temperature, 2);
  Serial.print("\t");
  Serial.print(dataRaw.humidity, 2);
  Serial.print("\t");
  Serial.print(dataRaw.stationPressure, 2);
  Serial.print("\t");
  Serial.print(dataRaw.seaLevelPressure, 2);
  Serial.print("\t");
  Serial.print(dataRaw.lightIntensity);
  Serial.print("\t");
  Serial.println(dataRaw.voltage, 2);

  // second line - smoothed data
  Serial.print("SMTH\t");
  Serial.print(dataSmooth.temperature, 2);
  Serial.print("\t");
  Serial.print(dataSmooth.humidity, 2);
  Serial.print("\t");
  Serial.print(dataSmooth.stationPressure, 2);
  Serial.print("\t");
  Serial.print(dataSmooth.seaLevelPressure, 2);
  Serial.println("");
  Serial.println("----------------------------------------------------");
} // printToSerialPort()

// *******************************************************
// print selected data to LCD display
void printToLCD(sensorData data, bool metricUnits)
{
  char temperatureSymbol = 'C';       // default is Celsius
  char pressureUnit[5] = "hPa ";          // default is hectoPascal
  if (metricUnits == false)           // user wants Imperial units
  {
    data.temperature = data.temperature * 1.8 + 32;
    data.seaLevelPressure = data.seaLevelPressure * HPA_TO_INHG;
    temperatureSymbol = 'F';
    strncpy(pressureUnit, "inHg", 5);
  }

  lcd.clear();
  lcd.print(data.temperature, 1);
  lcd.print((char)0xDF);              // prints degree symbol on LCD
  lcd.print(temperatureSymbol);
  lcd.print(' ');
  lcd.print(data.humidity, 0);
  lcd.print("% ");
  lcd.print(data.lightIntensity);

  // move to next line
  lcd.setCursor(0, 1);
  lcd.print(data.seaLevelPressure, 1);
  lcd.print(pressureUnit);
} // printToLCD()

// *******************************************************
// calculate dewpoint temperature in Celsius
// reference: http://en.wikipedia.org/wiki/Dew_point
float dewPoint(float celsius, float humidity)
{
  float b = 17.271;
  float c = 237.7;
  // log(x) =  conventional ln(x), log10(x) = conventional log(x)
  float gamma = (b * celsius) / (c + celsius) + log(humidity * 0.01);
  float Td = (c * gamma) / (b - gamma);
  return Td;
} // dewPoint()

// *******************************************************
// calculate relative sealevel pressure from absolute station pressure in hPa
// temperature in °C, elevation in m
// http://www.sandhurstweather.org.uk/barometric.pdf
// http://keisan.casio.com/exec/system/1224575267
float calculateSeaLevelPressure(float celsius, float stationPressure, float elevation)
{
  float slP = stationPressure / pow(2.718281828, -(elevation / ((273.15 + celsius) * 29.263)));
  return slP;
} // calculateSeaLevelPressure()

// *******************************************************
// *********************** END ***************************
// *******************************************************
